import h5py
import torch
from collections import defaultdict

class NetworkLoader:
    spec = None
    plural = None
    columns = None

    def __init__(self, june_world_path):
        self.june_world_path = june_world_path

    def _get_people_per_group(self):
        ret = defaultdict(lambda: [])
        with h5py.File(self.june_world_path, "r") as f:
            for column in self.columns:
                group_ids = f["population"]["group_ids"][:, column]
                group_specs = f["population"]["group_specs"][:, column]
                for i, (group_id, group_spec) in enumerate(zip(group_ids, group_specs)):
                    if group_spec.decode() != self.spec:
                        continue
                    ret[group_id].append(i)
        return ret

    def _get_group_ids(self):
        with h5py.File(self.june_world_path, "r") as f:
            group_ids = f[self.plural]["id"][:]
        return group_ids

    def load_network(self, data):
        people_per_group = self._get_people_per_group()
        adjlist_i = []
        adjlist_j = []
        for group_id, people in people_per_group.items():
            for person in people:
                adjlist_i.append(person)
                adjlist_j.append(group_id)
        data[self.spec].id = self._get_group_ids()
        data[self.spec].people = torch.tensor(
            [len(people_per_group[i]) for i in data[self.spec].id]
        )
        edge_type = ("agent", f"attends_{self.spec}", self.spec)
        new_edges = torch.vstack((torch.tensor(adjlist_i), torch.tensor(adjlist_j)))
        data[edge_type].edge_index = new_edges



from .base import (
    InfectionNetwork,
    InfectionNetworks,
    SchoolNetwork,
    CompanyNetwork,
    HouseholdNetwork,
    CareHomeNetwork,
    UniversityNetwork,
)
from .leisure_network import (
    PubNetwork,
    GroceryNetwork,
    CinemaNetwork,
    VisitNetwork,
    GymNetwork,
    CareVisitNetwork
)



import torch
import yaml

from grad_june.paths import default_config_path
from grad_june.utils import parse_distribution


class TransmissionSampler:
    def __init__(self, max_infectiousness, shape, rate, shift):
        self.max_infectiousness = max_infectiousness
        self.shape = shape
        self.rate = rate
        self.shift = shift

    def __call__(self, n):
        maxi = self.max_infectiousness.rsample((n,))
        shape = self.shape.rsample((n,))
        rate = self.rate.rsample((n,))
        shift = self.shift.rsample((n,))
        return torch.vstack((maxi, shape, rate, shift))

    @classmethod
    def from_file(cls, fpath=default_config_path):
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    @classmethod
    def from_parameters(cls, params):
        ret = {}
        tparams = params["transmission"]
        device = params["system"]["device"]
        for key in tparams:
            ret[key] = parse_distribution(tparams[key], device=device)
        return cls(**ret)


class TransmissionUpdater(torch.nn.Module):
    def forward(self, data, timer):
        shape = data["agent"]["infection_parameters"]["shape"]
        shift = data["agent"]["infection_parameters"]["shift"]
        rate = data["agent"]["infection_parameters"]["rate"]
        max_infectiousness = data["agent"]["infection_parameters"]["max_infectiousness"]
        time_from_infection = timer.now - data["agent"].infection_time
        sign = (torch.sign(time_from_infection - shift + 1e-10) + 1) / 2
        aux = torch.exp(-torch.lgamma(shape)) * torch.pow(
            (time_from_infection - shift) * rate, shape - 1.0
        )
        aux2 = torch.exp((shift - time_from_infection) * rate) * rate
        ret = max_infectiousness * sign * aux * aux2 * data["agent"].is_infected
        return ret



import torch

from grad_june.policies.policies import PolicyCollection, Policy


class Quarantine(Policy):
    spec = "quarantine"

    def __init__(self, start_date, end_date, stage_threshold, device):
        super().__init__(start_date=start_date, end_date=end_date, device=device)
        self.stage_threshold = stage_threshold

    def apply(self, symptom_stages, timer):
        if self.is_active(timer.date):
            ret = (symptom_stages < self.stage_threshold).to(torch.float)
        else:
            ret = torch.ones(symptom_stages.shape, device=symptom_stages.device)
        return ret


class QuarantinePolicies(PolicyCollection):
    def __init__(self, policies):
        super().__init__(policies)
        self.quarantine_mask = 1.0

    def apply(self, symptom_stages, timer):
        self.quarantine_mask = torch.ones(
            symptom_stages.shape, device=symptom_stages.device
        )
        for policy in self.policies:
            self.quarantine_mask = self.quarantine_mask * policy.apply(
                symptom_stages=symptom_stages, timer=timer
            )



from abc import ABC
import yaml
import re
import datetime
import torch

import grad_june
from grad_june.utils import read_date
from grad_june.paths import default_config_path


class Policy(torch.nn.Module):
    def __init__(self, start_date, end_date, device):
        super().__init__()
        self.start_date = read_date(start_date)
        self.end_date = read_date(end_date)
        self.device = device

    def apply(self):
        raise NotImplementedError

    def is_active(self, date: datetime.datetime) -> bool:
        """
        Returns true if the policy is active, false otherwise

        Parameters
        ----------
        date:
            date to check
        """
        return self.start_date <= date < self.end_date


class PolicyCollection(torch.nn.Module):
    def __init__(self, policies: Policy):
        """
        A collection of like policies active on the same date
        """
        super().__init__()
        self.policies = torch.nn.ModuleList(policies)

    def __getitem__(self, idx):
        return self.policies[idx]


class Policies(torch.nn.Module):
    def __init__(
        self,
        interaction_policies=None,
        quarantine_policies=None,
        close_venue_policies=None,
    ):
        super().__init__()
        self.interaction_policies = interaction_policies
        self.quarantine_policies = quarantine_policies
        self.close_venue_policies = close_venue_policies

    @classmethod
    def from_policy_list(cls, policies):
        if policies is None:
            policies = torch.nn.ModuleList([])
        from grad_june.policies import (
            InteractionPolicies,
            QuarantinePolicies,
            CloseVenuePolicies,
        )

        interaction_policies = InteractionPolicies(
            cls._get_policies_by_type(policies, "interaction")
        )
        quarantine_policies = QuarantinePolicies(
            cls._get_policies_by_type(policies, "quarantine")
        )
        close_venue_policies = CloseVenuePolicies(
            cls._get_policies_by_type(policies, "close_venue")
        )
        return cls(
            interaction_policies=interaction_policies,
            quarantine_policies=quarantine_policies,
            close_venue_policies=close_venue_policies,
        )

    @classmethod
    def from_file(cls, fpath=default_config_path):
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    @classmethod
    def from_parameters(cls, params):
        policy_params = params.get("policies", {})
        device = params["system"]["device"]
        policies = []
        for policy_collection in policy_params.values():
            for policy_name, policy_config in policy_collection.items():
                policies += cls._parse_policy_config(
                    policy_config, name=policy_name, device=device
                )
        return cls.from_policy_list(policies)

    @staticmethod
    def _parse_policy_config(config, name, device):
        camel_case_key = "".join(x.capitalize() or "_" for x in name.split("_"))
        policies = []
        policy_class = getattr(grad_june.policies, camel_case_key)
        if "start_date" not in config:
            for policy_i, policy_data_i in config.items():
                if (
                    "start_date" not in policy_data_i.keys()
                    or "end_date" not in policy_data_i.keys()
                ):
                    raise ValueError("policy config file not valid.")
                policies.append(policy_class(**policy_data_i, device=device))
        else:
            policies.append(policy_class(**config, device=device))
        return policies

    @classmethod
    def _get_policies_by_type(cls, policies, type):
        return [policy for policy in policies if policy.spec == type]

    def apply(self, data, timer):
        if self.quarantine_policies:
            self.quarantine_policies.apply(
                timer=timer, symptom_stages=data["agent"]["symptoms"]["current_stage"]
            )



import re
import yaml
import torch
from torch.utils.checkpoint import checkpoint
from torch_geometric.nn.conv import MessagePassing

from grad_june.paths import default_config_path
import grad_june.infection_networks


class InfectionNetwork(MessagePassing):
    def __init__(self, log_beta, device="cpu"):
        super().__init__( aggr="add", node_dim=-1)
        self.device = device
        if type(log_beta) != torch.nn.Parameter:
            self.log_beta = torch.tensor(float(log_beta))
        else:
            self.log_beta = log_beta
        self.name = self._get_name()

    @classmethod
    def from_parameters(cls, params):
        device = params["system"]["device"]
        return cls(device=device, **params["networks"][cls._get_name()])

    @classmethod
    def _get_name(cls):
        return "_".join(re.findall("[A-Z][^A-Z]*", cls.__name__)[:-1]).lower()

    def _get_edge_index(self, data):
        return data["attends_" + self.name].edge_index

    def _get_reverse_edge_index(self, data):
        return data["rev_attends_" + self.name].edge_index

    def _get_beta(self, policies, timer, data):
        interaction_policies = policies.interaction_policies
        beta = 10.0**self.log_beta
        if interaction_policies:
            beta = interaction_policies.apply(beta=beta, name=self.name, timer=timer)
        beta = beta * torch.ones(len(data[self.name]["id"]), device=self.device)
        return beta

    def _get_people_per_group(self, data):
        return data[self.name]["people"]

    def _get_transmissions(self, data, policies, timer):
        if policies.quarantine_policies:
            mask = policies.quarantine_policies.quarantine_mask
        else:
            mask = 1.0
        return mask * data["agent"].transmission

    def _get_susceptibilities(self, data, policies, timer):
        if policies.quarantine_policies:
            mask = policies.quarantine_policies.quarantine_mask
        else:
            mask = 1.0
        return mask * data["agent"].susceptibility

    def forward(self, data, timer, policies):
        beta = self._get_beta(policies=policies, timer=timer, data=data)
        people_per_group = self._get_people_per_group(data)
        p_contact = torch.maximum(
            torch.minimum(
                1.0 / (people_per_group - 1), torch.tensor(1.0, device=self.device)
            ),
            torch.tensor(0.0, device=self.device),
        )  # assumes constant n of contacts, change this in the future
        beta = beta * p_contact
        # remove people who are not really in this group
        transmissions = self._get_transmissions(
            data=data, policies=policies, timer=timer
        )
        susceptibilities = self._get_susceptibilities(
            data=data, policies=policies, timer=timer
        )
        edge_index = self._get_edge_index(data)
        cumulative_trans = self.propagate(edge_index, x=transmissions, y=beta)
        rev_edge_index = self._get_reverse_edge_index(data)
        trans_susc = self.propagate(
            rev_edge_index, x=cumulative_trans, y=susceptibilities
        )
        return trans_susc

    def message(self, x_j, y_i):
        return x_j * y_i


class InfectionNetworks(torch.nn.Module):
    def __init__(self, device="cpu", **kwargs):
        super().__init__()
        self.networks = torch.nn.ModuleDict(kwargs)
        self.device = device

    def __getitem__(self, item):
        return self.networks[item]

    @classmethod
    def from_parameters(cls, params):
        device = params["system"]["device"]
        network_params = params["networks"]
        network_dict = {}
        for key in network_params:
            network_name = "".join(word.title() for word in key.split("_"))
            network_name = network_name + "Network"
            network_class = getattr(grad_june.infection_networks, network_name)
            network = network_class.from_parameters(params)
            network_dict[key] = network
        return cls(device=device, **network_dict)

    @classmethod
    def from_file(cls, fpath=default_config_path):
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    def forward(
        self,
        data,
        timer,
        policies,
    ):
        n_agents = len(data["agent"].id)
        delta_time = timer.duration
        policies.apply(timer=timer, data=data)
        trans_susc = torch.zeros(n_agents, device=self.device)
        activity_order = timer.get_activity_order()
        if policies.close_venue_policies:
            activity_order = policies.close_venue_policies.apply(
                edge_types=activity_order, timer=timer
            )
        for activity in activity_order:
            network = self.networks[activity]
            trans_susc += network(data=data, timer=timer, policies=policies)
        trans_susc = torch.clamp(
            trans_susc, min=1e-6, max = 100
        )  # this is necessary to avoid gradient nans
        not_infected_probs = torch.exp(-trans_susc * delta_time)
        not_infected_probs = torch.clamp(not_infected_probs, min=0.0, max=1.0)
        return not_infected_probs


class HouseholdNetwork(InfectionNetwork):
    def _get_transmissions(self, data, policies, timer):
        return data["agent"].transmission

    def _get_susceptibilities(self, data, policies, timer):
        return data["agent"].susceptibility

    pass


class CareHomeNetwork(InfectionNetwork):
    pass


class SchoolNetwork(InfectionNetwork):
    pass


class CompanyNetwork(InfectionNetwork):
    pass


class UniversityNetwork(InfectionNetwork):
    pass




import numpy as np
import torch
from pathlib import Path
from itertools import chain
from torch import distributions as dist
import datetime
import random
from typing import Union
from copy import deepcopy
from torch_geometric.data import HeteroData
import torch_geometric.transforms as T

from grad_june.paths import grad_june_path
from grad_june.infection import infect_people_at_indices


def read_path(path_str):
    path = Path(path_str)
    if path.parts[0] == "@grad_june":
        path = Path("/".join(path_str.split("/")[1:]))
        path = grad_june_path / path
    return path


def read_date(date: Union[str, datetime.datetime]) -> datetime.datetime:
    """
    Read date in two possible formats, either string or datetime.date, both
    are translated into datetime.datetime to be used by the simulator

    Parameters
    ----------
    date:
        date to translate into datetime.datetime

    Returns
    -------
        date in datetime format
    """
    if type(date) is str:
        return datetime.datetime.strptime(date, "%Y-%m-%d")
    elif isinstance(date, datetime.date):
        return datetime.datetime.combine(date, datetime.datetime.min.time())
    else:
        raise TypeError("date must be a string or a datetime.date object")


def parse_age_probabilities(age_dict, fill_value=0):
    """
    Parses the age probability dictionaries into an array.
    """
    bins = []
    probabilities = []
    for age_range in age_dict:
        age_range_split = age_range.split("-")
        bins.append(int(age_range_split[0]))
        bins.append(int(age_range_split[1]))
        probabilities.append(age_dict[age_range])
    sorting_idx = np.argsort(bins[::2])
    bins = list(
        chain.from_iterable([bins[2 * idx], bins[2 * idx + 1]] for idx in sorting_idx)
    )
    probabilities = np.array(probabilities)[sorting_idx]
    probabilities_binned = []
    for prob in probabilities:
        probabilities_binned.append(fill_value)
        probabilities_binned.append(prob)
    probabilities_binned.append(fill_value)
    probabilities_per_age = []
    for age in range(100):
        idx = np.searchsorted(bins, age + 1)  # we do +1 to include the lower boundary
        probabilities_per_age.append(probabilities_binned[idx])
    return probabilities_per_age


def parse_distribution(dict, device):
    dd = deepcopy(dict)
    dist_name = dd.pop("dist")
    dist_class = getattr(dist, dist_name)
    input = {
        key: torch.tensor(value, device=device, dtype=torch.float)
        for key, value in dd.items()
    }
    return dist_class(**input)


def fix_seed(seed=None):
    if seed is None:
        seed = np.random.randint(0, 1000)
    print(f"Fixing seed to {seed}")
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def create_simple_connected_graph(n_agents):
    # avoid circular import
    from grad_june.transmission import TransmissionSampler

    data = HeteroData()
    sampler = TransmissionSampler.from_file()
    data["agent"].id = torch.arange(0, n_agents)
    data["agent"].age = torch.randint(0, 100, (n_agents,))
    data["agent"].sex = torch.randint(0, 2, (n_agents,))
    inf_params = {}
    inf_params_values = sampler(n_agents)
    inf_params["max_infectiousness"] = inf_params_values[0]
    inf_params["shape"] = inf_params_values[1]
    inf_params["rate"] = inf_params_values[2]
    inf_params["shift"] = inf_params_values[3]
    data["agent"].infection_parameters = inf_params
    data["agent"].transmission = torch.zeros(n_agents)
    data["agent"].susceptibility = torch.ones(n_agents)
    data["agent"].is_infected = torch.zeros(n_agents)
    data["agent"].infection_time = torch.zeros(n_agents)
    symptoms = {}
    symptoms["current_stage"] = torch.ones(n_agents, dtype=torch.long)
    symptoms["next_stage"] = torch.ones(n_agents, dtype=torch.long)
    symptoms["time_to_next_stage"] = torch.zeros(n_agents)
    data["agent"].symptoms = symptoms
    data["household"].id = torch.zeros(1)
    data["school"].id = torch.zeros(1)
    data["household"].people = torch.tensor([n_agents])
    data["school"].people = torch.tensor([n_agents])
    data["agent", "attends_household", "household"].edge_index = torch.vstack(
        (data["agent"].id[::2], torch.zeros(n_agents//2, dtype=torch.long))
    )
    data["agent", "attends_school", "school"].edge_index = torch.vstack(
        (data["agent"].id[1::2], torch.zeros(n_agents//2, dtype=torch.long))
    )
    data = T.ToUndirected()(data)
    return data



from operator import ne
import torch
import yaml

from grad_june.utils import parse_age_probabilities, parse_distribution
from grad_june.default_parameters import make_parameters
from grad_june.paths import default_config_path


class SymptomsSampler:
    def __init__(
        self,
        stages,
        stage_transition_probabilities,
        stage_transition_times,
        recovery_times,
        device,
    ):
        self.stages = stages
        self.stages_ids = torch.arange(0, len(stages))

        self.stage_transition_probabilities = (
            self._parse_stage_transition_probabilities(
                stage_transition_probabilities, device=device
            )
        )
        self.stage_transition_times = self._parse_stage_times(
            stage_transition_times, device=device
        )
        self.recovery_times = self._parse_stage_times(recovery_times, device=device)

    @classmethod
    def from_file(cls, fpath=default_config_path):
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    @classmethod
    def from_parameters(cls, params):
        return cls(**params["symptoms"], device=params["system"]["device"])

    def _parse_stage_transition_probabilities(
        self, stage_transition_probabilities, device
    ):
        ret = torch.zeros((len(self.stages), 100), device=device)
        for i, stage in enumerate(self.stages):
            if stage not in stage_transition_probabilities:
                continue
            ret[i] = torch.tensor(
                parse_age_probabilities(stage_transition_probabilities[stage]),
                dtype=torch.float32,
                device=device,
            )
        return ret

    def _parse_stage_times(self, stage_times, device):
        ret = {}
        for i, stage in enumerate(self.stages):
            if stage not in stage_times:
                ret[i] = None
            else:
                ret[i] = parse_distribution(stage_times[stage], device)
        return ret

    def _get_need_to_transition(self, current_stage, time_to_next_stage, time):
        """
        Gets a mask that is 1 for the agents that need to transition stage and 0
        for the ones that don't.
        """
        mask1 = time >= time_to_next_stage
        mask2 = current_stage < len(self.stages) - 1
        return mask1 * mask2

    def _get_prob_next_symptoms_stage(self, ages, stages):
        """
        Gets the probability of transitioning to the next symptoms stage,
        given the agents' ages and current stages.
        """
        probs = self.stage_transition_probabilities[stages, ages]
        return probs

    def sample_next_stage(
        self, ages, current_stage, next_stage, time_to_next_stage, time
    ):
        """
        Samples the next stage for each agent that needs updating.
        Returns the current stages, next stages, and the time to next stages.
        """
        # Check who has reached stage completion time and move them forward
        mask_transition = self._get_need_to_transition(
            current_stage, time_to_next_stage, time
        )
        n_agents = ages.shape[0]
        current_stage = current_stage - (current_stage - next_stage) * mask_transition
        # Sample possible next stages
        probs = self._get_prob_next_symptoms_stage(ages, current_stage.long())
        mask_symp_stage = torch.bernoulli(probs).to(torch.bool) # no dependence on parameters here.
        # These ones would recover
        mask_recovered_stage = ~mask_symp_stage
        for i in range(
            2, len(self.stages) - 1
        ):  # skip recovered, susceptible, and dead
            # Check people at this stage that need updating
            mask_stage = (current_stage == i)
            # this makes sure the gradient flows, since we lost it in the previous line.
            mask_stage = mask_stage * current_stage / i 
            mask_updating = mask_stage * mask_transition

            # These people progress to another disease stage
            mask_symp = mask_updating * mask_symp_stage
            n_symp = mask_symp.max()  # faster than sum
            if n_symp > 0:
                next_stage = next_stage + mask_symp
                time_to_next_stage = (
                    time_to_next_stage
                    + self.stage_transition_times[i].rsample((n_agents,)) * mask_symp
                )

            # These people will recover
            mask_rec = mask_updating * mask_recovered_stage
            n_rec = mask_rec.max()
            if n_rec > 0:
                next_stage = next_stage - next_stage * mask_rec  # Set to 0
                time_to_next_stage = (
                    time_to_next_stage
                    + self.recovery_times[i].rsample((n_agents,)) * mask_rec
                )
        return current_stage, next_stage, time_to_next_stage


class SymptomsUpdater(torch.nn.Module):
    """
    A PyTorch module for updating symptoms of agents in an epidemic simulation.

    Args:
        symptoms_sampler (SymptomsSampler): A sampler for generating new symptoms based on current symptoms and other factors.

    Attributes:
        symptoms_sampler (SymptomsSampler): A sampler for generating new symptoms based on current symptoms and other factors.

    Methods:
        from_file(cls, fpath=default_config_path): Class method for creating a SymptomsUpdater instance from a YAML file.
        from_parameters(cls, params): Class method for creating a SymptomsUpdater instance from a dictionary of parameters.
        forward(self, data, timer, new_infected): Method for updating symptoms of agents based on current symptoms and other factors.
        stages_ids(self): Property for getting the IDs of all possible symptom stages.

    Raises:
        TypeError: If symptoms_sampler is not an instance of SymptomsSampler.
        KeyError: If data does not contain the "agent" key.
        KeyError: If symptoms does not contain the "current_stage", "next_stage", or "time_to_next_stage" keys.

    Returns:
        A SymptomsUpdater instance.
    """

    def __init__(self, symptoms_sampler):
        super().__init__()
        if not isinstance(symptoms_sampler, SymptomsSampler):
            raise TypeError("symptoms_sampler must be an instance of SymptomsSampler.")
        self.symptoms_sampler = symptoms_sampler

    @classmethod
    def from_file(cls, fpath=default_config_path):
        """
        Class method for creating a SymptomsUpdater instance from a YAML file.

        Args:
            fpath (str): The path to the YAML file containing the parameters.

        Raises:
            FileNotFoundError: If the file at fpath does not exist.
            yaml.YAMLError: If the file at fpath is not a valid YAML file.
            KeyError: If the YAML file does not contain the "symptoms_sampler" key.

        Returns:
            A SymptomsUpdater instance.
        """
        try:
            with open(fpath, "r") as f:
                params = yaml.safe_load(f)
        except FileNotFoundError:
            raise FileNotFoundError(f"No file found at {fpath}.")
        except yaml.YAMLError:
            raise yaml.YAMLError(f"Invalid YAML file at {fpath}.")
        ss = SymptomsSampler.from_parameters(params)
        return cls(symptoms_sampler=ss)

    @classmethod
    def from_parameters(cls, params):
        """
        Class method for creating a SymptomsUpdater instance from a dictionary of parameters.

        Args:
            params (dict): A dictionary of parameters.

        Raises:

        Returns:
            A SymptomsUpdater instance.
        """
        ss = SymptomsSampler.from_parameters(params)
        return cls(symptoms_sampler=ss)

    def forward(self, data, timer, new_infected):
        """
        Method for updating symptoms of agents based on current symptoms and other factors.

        Args:
            data (dict): A dictionary containing information about the agents.
            timer (Timer): A timer object for keeping track of time in the simulation.
            new_infected (int): The number of new agents who have become infected.

        Raises:
            KeyError: If data does not contain the "agent" key.
            KeyError: If symptoms does not contain the "current_stage", "next_stage", or "time_to_next_stage" keys.

        Returns:
            A dictionary containing the updated symptoms of the agents.
        """
        try:
            symptoms = data["agent"].symptoms
        except KeyError:
            raise KeyError("data must contain the 'agent' key.")
        if not all(key in symptoms for key in ["current_stage", "next_stage", "time_to_next_stage"]):
            raise KeyError("symptoms must contain the 'current_stage', 'next_stage', and 'time_to_next_stage' keys.")
        time = timer.now
        symptoms["next_stage"] = symptoms["next_stage"] + new_infected * (
            2.0 - symptoms["next_stage"]
        )
        symptoms["time_to_next_stage"] = symptoms[
            "time_to_next_stage"
        ] + new_infected * (time - symptoms["time_to_next_stage"])
        (
            current_stage,
            next_stage,
            time_to_next_stage,
        ) = self.symptoms_sampler.sample_next_stage(
            ages=data["agent"].age,
            current_stage=symptoms["current_stage"],
            next_stage=symptoms["next_stage"],
            time_to_next_stage=symptoms["time_to_next_stage"],
            time=time,
        )
        symptoms["current_stage"] = current_stage
        symptoms["next_stage"] = next_stage
        symptoms["time_to_next_stage"] = time_to_next_stage
        return symptoms

    @property
    def stages_ids(self):
        """
        Property for getting the IDs of all possible symptom stages.

        Returns:
            A list of integers representing the IDs of all possible symptom stages.
        """
        return self.symptoms_sampler.stages_ids



import torch

from grad_june.infection_networks import InfectionNetwork
from grad_june.utils import parse_age_probabilities


class LeisureNetwork(InfectionNetwork):
    def __init__(self, log_beta, leisure_probabilities, device):
        super().__init__(log_beta=log_beta, device=device)
        self.leisure_probabilities = self._parse_leisure_probabilities(
            leisure_probabilities
        )
        self.weekday_probabilities = None
        self.weekend_probabilities = None

    @classmethod
    def from_parameters(cls, params):
        device = params["system"]["device"]
        leisure_probabilities = params["leisure"][cls._get_name()]
        return cls(
            device=device,
            leisure_probabilities=leisure_probabilities,
            **params["networks"][cls._get_name()]
        )

    def _parse_leisure_probabilities(self, leisure_probabilities):
        ret = torch.zeros((2, 2, 100), device=self.device)
        for i, day_type in enumerate(["weekday", "weekend"]):
            for j, sex in enumerate(["male", "female"]):
                parsed_probs = parse_age_probabilities(
                    leisure_probabilities[day_type][sex]
                )
                ret[i, j, :] = torch.tensor(parsed_probs, device=self.device)
        return ret

    def initialize_leisure_probabilities(self, data):
        self.weekday_probabilities = self.leisure_probabilities[
            0, data["agent"].sex, data["agent"].age
        ]
        self.weekend_probabilities = self.leisure_probabilities[
            1, data["agent"].sex, data["agent"].age
        ]

    def _get_edge_index(self, data):
        return data["attends_leisure"].edge_index

    def _get_reverse_edge_index(self, data):
        return data["rev_attends_leisure"].edge_index

    def _get_beta(self, policies, timer, data):
        interaction_policies = policies.interaction_policies
        beta = 10.0**self.log_beta
        if interaction_policies:
            beta = interaction_policies.apply(beta=beta, name=self.name, timer=timer)
        beta = beta * torch.ones(len(data["leisure"]["id"]), device=self.device)
        return beta

    def _get_people_per_group(self, data):
        return data["leisure"]["people"]

    def _get_transmissions(self, data, policies, timer):
        if self.weekday_probabilities is None:
            self.initialize_leisure_probabilities(data)
        if policies.quarantine_policies:
            mask = policies.quarantine_policies.quarantine_mask
        else:
            mask = 1.0
        if timer.day_type == "weekday":
            leisure_mask = self.weekday_probabilities
        else:
            leisure_mask = self.weekend_probabilities
        return mask * leisure_mask * data["agent"].transmission

    def _get_susceptibilities(self, data, policies, timer):
        if self.weekday_probabilities is None:
            self.initialize_leisure_probabilities(data)
        if policies.quarantine_policies:
            mask = policies.quarantine_policies.quarantine_mask
        else:
            mask = 1.0
        if timer.day_type == "weekday":
            leisure_mask = self.weekday_probabilities
        else:
            leisure_mask = self.weekend_probabilities
        return mask * leisure_mask * data["agent"].susceptibility


class PubNetwork(LeisureNetwork):
    pass


class CinemaNetwork(LeisureNetwork):
    pass


class GroceryNetwork(LeisureNetwork):
    pass


class GymNetwork(LeisureNetwork):
    pass


class VisitNetwork(LeisureNetwork):
    pass

class CareVisitNetwork(LeisureNetwork):
    def _get_susceptibilities(self, data, policies, timer):
        mask_age = data["agent"].age > 75
        if self.weekday_probabilities is None:
            self.initialize_leisure_probabilities(data)
        if policies.quarantine_policies:
            mask = policies.quarantine_policies.quarantine_mask
        else:
            mask = 1.0
        if timer.day_type == "weekday":
            leisure_mask = self.weekday_probabilities
        else:
            leisure_mask = self.weekend_probabilities
        return mask * leisure_mask * data["agent"].susceptibility * mask_age



import subprocess
import torch


def get_gpu_memory_map():
    """Get the current gpu usage.

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """
    result = subprocess.check_output(
        ["nvidia-smi", "--query-gpu=memory.used", "--format=csv,nounits,noheader"],
        encoding="utf-8",
    )
    # Convert lines into a dictionary
    gpu_memory = [int(x) for x in result.strip().split("\n")]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map


def get_fraction_gpu_used(i):
    total = torch.cuda.get_device_properties(f"cuda:{i}").total_memory / 1e6  # MB
    used = get_gpu_memory_map()[i]
    return used / total



from grad_june.policies import Policy, PolicyCollection


class CloseVenue(Policy):
    spec = "close_venue"

    def __init__(self, start_date, end_date, names, device):
        super().__init__(start_date=start_date, end_date=end_date, device=device)
        self.edge_type_to_close = set([f"{name}" for name in names])

    def apply(self, edge_types, timer):
        if self.is_active(timer.date):
            return [edge for edge in edge_types if edge not in self.edge_type_to_close]
        else:
            return edge_types


class CloseVenuePolicies(PolicyCollection):
    def apply(self, edge_types, timer):
        for policy in self.policies:
            edge_types = policy.apply(edge_types=edge_types, timer=timer)
        return edge_types



import torch

from grad_june.policies import Policy, PolicyCollection


class InteractionPolicy(Policy):
    spec = "interaction"


class InteractionPolicies(PolicyCollection):
    def apply(self, beta, name, timer):
        for policy in self.policies:
            beta = policy.apply(beta=beta, name=name, timer=timer)
        return beta


class SocialDistancing(InteractionPolicy):
    def __init__(self, start_date, end_date, beta_factors, device):
        super().__init__(start_date=start_date, end_date=end_date, device=device)
        beta_factors_ = {}
        for key in beta_factors:
            beta_factors_[key] = torch.tensor(float(beta_factors[key]), device=device)
        self.beta_factors = beta_factors_

    def apply(self, beta, name, timer):
        if self.is_active(timer.date):
            factor = self.beta_factors.get(name, self.beta_factors.get("all", torch.tensor(1.0)))
            #print(f"factor for {name} is {factor}")
            return beta * factor
        else:
            return beta



import torch_geometric.transforms as T

from grad_june.june_world_loader.household_loader import HouseholdNetworkLoader
from grad_june.june_world_loader.care_home_loader import CareHomeNetworkLoader
from grad_june.june_world_loader.company_loader import CompanyNetworkLoader
from grad_june.june_world_loader.school_loader import SchoolNetworkLoader 
from grad_june.june_world_loader.university_loader import UniversityNetworkLoader
from grad_june.june_world_loader.leisure_loader import LeisureNetworkLoader 


class GraphLoader:
    def __init__(self, june_world_path, k_leisure=3):
        self.june_world_path = june_world_path
        self.k_leisure = k_leisure

    def load_graph(
        self,
        data,
        load_leisure=True,
        loaders=(
            HouseholdNetworkLoader,
            CareHomeNetworkLoader,
            CompanyNetworkLoader,
            SchoolNetworkLoader,
            UniversityNetworkLoader,
        ),
    ):
        for network_loader_class in loaders:
            print(f"Loading {network_loader_class}...")
            network_loader = network_loader_class(self.june_world_path)
            network_loader.load_network(data)
        if load_leisure:
            print("Loading leisure ...")
            leisure_loader = LeisureNetworkLoader(
                self.june_world_path, k=self.k_leisure
            )
            leisure_loader.load_network(data)
        data = T.ToUndirected()(data)
        return data



from collections import defaultdict
import numpy as np
import h5py
import torch

from sklearn.neighbors import BallTree

class LeisureNetworkLoader:
    def __init__(self, june_world_path, k=1):
        self.june_world_path = june_world_path
        self._super_area_coordinates = self._get_super_area_coordinates()
        self._super_area_ids = self._get_super_area_ids()
        self._ball_tree = self._generate_ball_tree()
        self.k = k

    def _get_super_area_coordinates(self):
        with h5py.File(self.june_world_path, "r") as f:
            super_area_coords = f["geography"]["super_area_coordinates"][:]
        super_area_coords = np.array(
            [np.deg2rad(coordinates) for coordinates in super_area_coords]
        )
        return super_area_coords

    def _get_super_area_ids(self):
        with h5py.File(self.june_world_path, "r") as f:
            super_area_ids = f["geography"]["super_area_id"][:]
        return super_area_ids

    def _get_people_per_super_area(self):
        ret = {}
        with h5py.File(self.june_world_path, "r") as f:
            people_super_area = f["population"]["super_area"][:]
            for super_area_id in self._super_area_ids:
                people_in_super_area = np.where(people_super_area == super_area_id)[0]
                ret[super_area_id] = list(people_in_super_area)
        return ret

    def _generate_ball_tree(self):
        ball_tree = BallTree(self._super_area_coordinates, metric="haversine")
        return ball_tree

    def _get_closest_super_areas(self, super_area, k=3):
        coordinates = self._super_area_coordinates[super_area]
        dist, ind = self._ball_tree.query(coordinates.reshape(1, -1), k=k)
        return ind[0]

    def _get_close_people_per_super_area(self, k):
        ret = {}
        people_per_super_area = self._get_people_per_super_area()
        for super_area in self._super_area_ids:
            closest = self._get_closest_super_areas(super_area, k=k)
            people = []
            for sa in closest:
                people += people_per_super_area[sa]
            ret[super_area] = people
        return ret

    def load_network(self, data):
        close_people_per_super_area = self._get_close_people_per_super_area(k=self.k)
        ret = torch.empty((2, 0), dtype=torch.long)
        for super_area in close_people_per_super_area:
            people = torch.tensor(
                close_people_per_super_area[super_area], dtype=torch.long
            )
            edges = torch.vstack(
                (people, super_area * torch.ones(len(people), dtype=torch.long))
            )
            ret = torch.hstack((ret, edges))
        data["agent", "attends_leisure", "leisure"].edge_index = ret
        data["leisure"].id = torch.tensor(list(close_people_per_super_area.keys()))
        data["leisure"].people = torch.tensor(
            [len(close_people_per_super_area[sa]) for sa in close_people_per_super_area]
        )

#    def load_network(self, data):
#        close_people_per_super_area = self._get_close_people_per_super_area(k=self.k)
#        ret = torch.empty((2, 0), dtype=torch.long)
#        for super_area in close_people_per_super_area:
#            to_choose = int(len(close_people_per_super_area[super_area]) / self.k)
#            indices = torch.randperm(len(close_people_per_super_area[super_area]))[
#                :to_choose
#            ]
#            people = torch.tensor(
#                close_people_per_super_area[super_area], dtype=torch.long
#            )[indices]
#            edges = torch.vstack(
#                (people, super_area * torch.ones(len(people), dtype=torch.long))
#            )
#            ret = torch.hstack((ret, edges))
#        data["agent", "attends_leisure", "leisure"].edge_index = ret
#        data["leisure"].id = torch.tensor(list(close_people_per_super_area.keys()))
#        data["leisure"].people = torch.tensor(
#            [len(close_people_per_super_area[sa]) for sa in close_people_per_super_area]
#        )
#
#



from pathlib import Path
import os

grad_june_path = Path(os.path.abspath(__file__)).parent.parent

default_config_path = (
    Path(os.path.abspath(__file__)).parent.parent / "configs/default.yaml"
)



"""
Taken from JUNE.
"""

import calendar
import datetime
import yaml
from typing import List

from grad_june.paths import default_config_path

SECONDS_PER_DAY = 24 * 60 * 60

activity_hierarchy = [
    "school",
    "university",
    "company",
    "care_home",
    "pub",
    "gym",
    "grocery",
    "visit",
    "care_visit",
    "cinema",
    "household",
]


class Timer:
    def __init__(
        self,
        initial_day: str = "2020-03-01",
        total_days: int = 10,
        weekday_step_duration: List[int] = (12, 12),
        weekend_step_duration: List[int] = (24,),
        weekday_activities: List[List[str]] = (
            ("school", "household"),
            ("pub", "household",),
        ),
        weekend_activities: List[List[str]] = (("household",),),
    ):

        self.initial_date = datetime.datetime(
            *[int(value) for value in initial_day.split("-")]
        )
        self.total_days = total_days
        self.weekday_step_duration = weekday_step_duration
        self.weekend_step_duration = weekend_step_duration
        self.weekday_activities = weekday_activities
        self.weekend_activities = weekend_activities

        self.previous_date = self.initial_date
        self.final_date = self.initial_date + datetime.timedelta(days=total_days)
        self.date = datetime.datetime(*[int(value) for value in initial_day.split("-")])
        self.shift = 0
        self.delta_time = datetime.timedelta(hours=self.shift_duration)
        self.n_timesteps = 0

    @classmethod
    def from_file(cls, fpath=default_config_path):
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    @classmethod
    def from_parameters(cls, params):
        time_config = params["timer"]
        return cls(
            initial_day=time_config["initial_day"],
            total_days=time_config["total_days"],
            weekday_step_duration=time_config["step_duration"]["weekday"],
            weekend_step_duration=time_config["step_duration"]["weekend"],
            weekday_activities=time_config["step_activities"]["weekday"],
            weekend_activities=time_config["step_activities"]["weekend"],
        )

    @property
    def is_weekend(self):
        week_number = self.date.weekday()
        if week_number < 5:
            return False
        return True

    @property
    def day_type(self):
        week_number = self.date.weekday()
        if week_number < 5:
            return "weekday"
        else:
            return "weekend"

    @property
    def now(self):
        difference = self.date - self.initial_date
        return difference.total_seconds() / SECONDS_PER_DAY

    @property
    def date_str(self):
        return self.date.date().strftime("%Y-%m-%d")

    @property
    def duration(self):
        return self.delta_time.total_seconds() / SECONDS_PER_DAY

    @property
    def day(self):
        return int(self.now)

    @property
    def day_of_week(self):
        return calendar.day_name[self.date.weekday()]

    @property
    def activities(self):
        type_day = "weekend" if self.is_weekend else "weekday"
        return getattr(self, type_day + "_activities")[self.shift]

    @property
    def shift_duration(self):
        type_day = "weekend" if self.is_weekend else "weekday"
        return getattr(self, type_day + "_step_duration")[self.shift]

    def reset(self):
        self.date = self.initial_date
        self.shift = 0
        self.delta_time = datetime.timedelta(hours=self.shift_duration)
        self.previous_date = self.initial_date

    def __next__(self):
        self.previous_date = self.date
        self.date += self.delta_time
        self.shift += 1
        if self.previous_date.day != self.date.day:
            self.shift = 0
        self.delta_time = datetime.timedelta(hours=self.shift_duration)
        self.n_timesteps += 1
        return self.date

    def _apply_activity_hierarchy(self, activities):
        """
        Returns a list of activities with the right order,
        obeying the permanent activity hierarcy and shuflling
        the random one.

        Parameters
        ----------
        activities:
            list of activities that take place at a given time step
        Returns
        -------
        Ordered list of activities according to hierarchy
        """
        activities.sort(key=lambda x: activity_hierarchy.index(x))
        return activities

    def get_activity_order(self):
        return self._apply_activity_hierarchy(list(self.activities))



import h5py
import torch
import numpy as np


class AgentDataLoader:
    def __init__(self, june_world_path):
        self.june_world_path = june_world_path

    def _get_socioeconomic_indices(self):
        bins = [0, 0.20, 0.4, 0.6, 0.8, 1.0]
        with h5py.File(self.june_world_path, "r") as f:
            agent_area_ids = f["population"]["area"][:]
            socio_indices = f["geography"]["area_socioeconomic_indices"][:][
                agent_area_ids
            ]
            socio_indices = np.digitize(socio_indices, bins)
        return torch.tensor(socio_indices, dtype=torch.int8)

    def load_agent_data(self, data):
        with h5py.File(self.june_world_path, "r") as f:
            population = f["population"]
            data["agent"].id = torch.tensor(population["id"][:])
            data["agent"].age = torch.tensor(population["age"][:])
            data["agent"].ethnicity = population["ethnicity"][:].astype("U")
            data["agent"].socioeconomic_index = self._get_socioeconomic_indices()
            area_ids = population["area"][:]
            area_names = f["geography"]["area_name"][:][area_ids].astype("U")
            data["agent"].area = area_names
            sexes = population["sex"][:].astype(str).astype(object)
            sexes[sexes == "m"] = 0
            sexes[sexes == "f"] = 1
            data["agent"].sex = torch.tensor(sexes.astype(int))



from .policies import Policy, Policies, PolicyCollection
from .interaction_policies import InteractionPolicies, SocialDistancing
from .close_venue_policies import CloseVenue, CloseVenuePolicies
from .quarantine_policies import Quarantine, QuarantinePolicies



import torch

class IsInfectedSampler(torch.nn.Module):
    def forward(self, not_infected_probs):
        """
        Here we need to sample the infection status of each agent and the variant that
        the agent gets in case of infection.
        To do this, we construct a tensor of size [M+1, N], where M is the number of
        variants and N the number of agents. The extra dimension in M will represent
        the agent not getting infected, so that it can be sampled as an outcome using
        the Gumbel-Softmax reparametrization of the categorical distribution.
        """
        logits = torch.vstack((not_infected_probs, 1.0 - not_infected_probs)).log()
        infection = torch.nn.functional.gumbel_softmax(
            logits, dim=0, tau=0.1, hard=True
        )
        is_infected = 1.0 - infection[0, :]
        return is_infected


def infect_people(data, timer, new_infected):
    data["agent"].susceptibility = torch.clamp(
        data["agent"].susceptibility - new_infected, min=0.0
    )
    data["agent"].is_infected = data["agent"].is_infected + new_infected
    data["agent"].infection_time = data["agent"].infection_time + new_infected * (
        timer.now - data["agent"].infection_time
    )


def infect_fraction_of_people(
    data, timer, symptoms_updater, fraction, device
):
    n_infections = data["agent"].susceptibility.shape[0]
    n_agents = data["agent"].id.shape[0]
    probs = fraction * torch.ones(n_agents, device=device)
    sampler = IsInfectedSampler()
    new_infected = sampler(
        1.0 - probs
    )  # sampler takes not inf probs
    infect_people(data, timer, new_infected)
    return new_infected


def infect_people_at_indices(data, indices, device="cpu"):
    susc = data["agent"]["susceptibility"].cpu().numpy()
    is_inf = data["agent"]["is_infected"].cpu().numpy()
    inf_t = data["agent"]["infection_time"].cpu().numpy()
    next_stage = data["agent"]["symptoms"]["next_stage"].cpu().numpy()
    current_stage = data["agent"]["symptoms"]["current_stage"].cpu().numpy()
    susc[indices] = 0.0
    is_inf[indices] = 1.0
    inf_t[indices] = 0.0
    next_stage[indices] = 2
    current_stage[indices] = 1
    data["agent"]["susceptibility"] = torch.tensor(susc, device=device)
    data["agent"]["is_infected"] = torch.tensor(is_inf, device=device)
    data["agent"]["infection_time"] = torch.tensor(inf_t, device=device)
    data["agent"]["symptoms"]["next_stage"] = torch.tensor(next_stage, device=device)
    data["agent"]["symptoms"]["current_stage"] = torch.tensor(
        current_stage, device=device
    )
    return data



import torch
import pickle
import numpy as np
import pandas as pd
from torch.utils.checkpoint import checkpoint
import yaml
from pathlib import Path

from grad_june.paths import default_config_path
from grad_june import GradJune, Timer, TransmissionSampler
from grad_june.utils import read_path
from grad_june.infection import infect_fraction_of_people


class Runner(torch.nn.Module):
    def __init__(
        self,
        model,
        data,
        timer,
        log_fraction_initial_cases,
        save_path,
        parameters,
        age_bins=(0, 18, 65, 100),
    ):
        super().__init__()
        self.model = model
        self.data = data
        self.data_backup = self.backup_infection_data(data)
        self.timer = timer
        self.log_fraction_initial_cases = log_fraction_initial_cases
        self.device = model.device
        self.age_bins = torch.tensor(age_bins, device=self.device)
        self.ethnicities = np.sort(np.unique(data["agent"].ethnicity))
        self.n_agents = data["agent"].id.shape[0]
        self.population_by_age = self.get_people_by_age()
        self.save_path = Path(save_path)
        self.input_parameters = parameters
        self.restore_initial_data()

    @classmethod
    def from_file(cls, fpath=default_config_path):
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    @classmethod
    def from_parameters(cls, params):
        model = GradJune.from_parameters(params)
        data = cls.get_data(params)
        timer = Timer.from_parameters(params)
        age_bins_to_save = params.get("age_bins_to_save", (0, 18, 65, 100))
        return cls(
            model=model,
            data=data,
            timer=timer,
            log_fraction_initial_cases=params["infection_seed"][
                "log_fraction_initial_cases"
            ],
            save_path=params["save_path"],
            parameters=params,
            age_bins = age_bins_to_save
        )

    @staticmethod
    def get_data(params):
        device = params["system"]["device"]
        data_path = read_path(params["data_path"])
        with open(data_path, "rb") as f:
            data = pickle.load(f).to(device)
        n_agents = len(data["agent"]["id"])
        inf_params = {}
        transmission_sampler = TransmissionSampler.from_parameters(params)
        transmission_values = transmission_sampler(n_agents)
        inf_params["max_infectiousness"] = transmission_values[0, :]
        inf_params["shape"] = transmission_values[1, :]
        inf_params["rate"] = transmission_values[2, :]
        inf_params["shift"] = transmission_values[3, :]
        data["agent"].infection_parameters = inf_params
        data["agent"].transmission = torch.zeros(n_agents, device=device)
        data["agent"].susceptibility = torch.ones(n_agents, device=device)
        data["agent"].is_infected = torch.zeros(n_agents, device=device)
        data["agent"].infection_time = torch.zeros(n_agents, device=device)
        symptoms = {}
        symptoms["current_stage"] = torch.ones(
            n_agents, dtype=torch.long, device=device
        )
        symptoms["next_stage"] = torch.ones(n_agents, dtype=torch.long, device=device)
        symptoms["time_to_next_stage"] = torch.zeros(n_agents, device=device)
        data["agent"].symptoms = symptoms
        return data

    def backup_infection_data(self, data):
        ret = {}
        ret["susceptibility"] = data["agent"].susceptibility.detach().clone()
        ret["is_infected"] = data["agent"].is_infected.detach().clone()
        ret["infection_time"] = data["agent"].infection_time.detach().clone()
        ret["transmission"] = data["agent"].transmission.detach().clone()
        symptoms = {}
        symptoms["current_stage"] = (
            data["agent"]["symptoms"]["current_stage"].detach().clone()
        )
        symptoms["next_stage"] = (
            data["agent"]["symptoms"]["next_stage"].detach().clone()
        )
        symptoms["time_to_next_stage"] = (
            data["agent"]["symptoms"]["time_to_next_stage"].detach().clone()
        )
        ret["symptoms"] = symptoms
        return ret

    def restore_initial_data(self):
        self.data["agent"].transmission = (
            self.data_backup["transmission"].detach().clone()
        )
        self.data["agent"].susceptibility = (
            self.data_backup["susceptibility"].detach().clone()
        )
        self.data["agent"].is_infected = (
            self.data_backup["is_infected"].detach().clone()
        )
        self.data["agent"].infection_time = (
            self.data_backup["infection_time"].detach().clone()
        )
        self.data["agent"].symptoms["current_stage"] = (
            self.data_backup["symptoms"]["current_stage"].detach().clone()
        )
        self.data["agent"].symptoms["next_stage"] = (
            self.data_backup["symptoms"]["next_stage"].detach().clone()
        )
        self.data["agent"].symptoms["time_to_next_stage"] = (
            self.data_backup["symptoms"]["time_to_next_stage"].detach().clone()
        )
        # reset results
        self.data["results"] = {}
        self.data["results"]["deaths_per_timestep"] = None

    def set_initial_cases(self):
        fraction_initial_cases = 10.0**self.log_fraction_initial_cases
        new_infected = infect_fraction_of_people(
            data=self.data,
            timer=self.timer,
            symptoms_updater=self.model.symptoms_updater,
            device=self.device,
            fraction=fraction_initial_cases,
        )
        self.model.symptoms_updater(
            data=self.data, timer=self.timer, new_infected=new_infected
        )

    def forward(self):
        timer = self.timer
        model = self.model
        data = self.data
        timer.reset()
        self.restore_initial_data()
        self.set_initial_cases()
        cases_per_timestep = data["agent"].is_infected.sum()
        cases_by_age = self.get_cases_by_age(data)
        self.store_differentiable_deaths(data)
        dates = [timer.date]
        i = 0
        while timer.date < timer.final_date:
            i += 1
            next(timer)
            data = model(data, timer)
            cases = data["agent"].is_infected.sum()
            cases_per_timestep = torch.hstack((cases_per_timestep, cases))
            self.store_differentiable_deaths(data)
            cases_age = self.get_cases_by_age(data)
            cases_by_age = torch.vstack((cases_by_age, cases_age))
            dates.append(timer.date)
        results = {
            "dates": dates,
            "cases_per_timestep": cases_per_timestep,
            "daily_cases_per_timestep": torch.diff(
                cases_per_timestep, prepend=torch.tensor([0.0], device=self.device)
            ),
            "deaths_per_timestep": data.results["deaths_per_timestep"],
        }
        for (i, key) in enumerate(self.age_bins[1:]):
            results[f"cases_by_age_{key:02d}"] = cases_by_age[:, i]
        return results, data["agent"].is_infected

    def save_results(self, results, is_infected):
        self.save_path.mkdir(exist_ok=True, parents=True)
        df = pd.DataFrame(index=results["dates"])
        df.index.name = "date"
        for key in results:
            if key in ("dates"):
                continue
            df[key] = results[key].detach().cpu().numpy()
        df.to_csv(self.save_path / "results.csv")
        df = pd.DataFrame()
        df["is_infected"] = is_infected
        df.to_csv(self.save_path / "results_is_infected.csv")

    def store_differentiable_deaths(self, data):
        """
        Returns differentiable deaths. The results are stored
        in data["results"]
        """
        symptoms = data["agent"].symptoms
        dead_idx = self.model.symptoms_updater.stages_ids[-1]
        deaths = (
            (symptoms["current_stage"] == dead_idx)
            * symptoms["current_stage"]
            / dead_idx
        )
        if data["results"]["deaths_per_timestep"] is not None:
            data["results"]["deaths_per_timestep"] = torch.hstack(
                (data["results"]["deaths_per_timestep"], deaths.sum())
            )
        else:
            data["results"]["deaths_per_timestep"] = deaths.sum()

    def get_cases_by_age(self, data):
        ret = torch.zeros(self.age_bins.shape[0] - 1, device=self.device)
        for i in range(1, self.age_bins.shape[0]):
            mask1 = data["agent"].age < self.age_bins[i]
            mask2 = data["agent"].age > self.age_bins[i - 1]
            mask = mask1 * mask2
            ret[i - 1] = (data["agent"].is_infected * mask).sum()
        return ret

    def get_people_by_age(self):
        ret = {}
        for i in range(1, self.age_bins.shape[0]):
            mask1 = self.data["agent"].age < self.age_bins[i]
            mask2 = self.data["agent"].age > self.age_bins[i - 1]
            mask = mask1 * mask2
            ret[int(self.age_bins[i].item())] = mask.sum()
        return ret

    def get_cases_by_ethnicity(self, data):
        ret = torch.zeros(len(self.ethnicities), device=self.device)
        for i, ethnicity in enumerate(self.ethnicities):
            mask = torch.tensor(
                self.data["agent"].ethnicity == ethnicity, device=self.device
            )
            ret[i] = (mask * data["agent"].is_infected).sum()
        return ret



from .june_world_loader import GraphLoader, AgentDataLoader
from .infection_networks import InfectionNetworks 
from .transmission import TransmissionUpdater, TransmissionSampler
from .symptoms import SymptomsUpdater
from .infection import IsInfectedSampler
from .model import GradJune
from .timer import Timer
from .policies import Policies
from .runner import Runner



from pathlib import Path
import torch
import torch_geometric.transforms as T
import numpy as np
import random
from pytest import fixture
from torch_geometric.data import HeteroData

from grad_june.transmission import TransmissionSampler
from grad_june.infection import infect_people_at_indices
from grad_june.timer import Timer


@fixture(autouse=True)
def set_random_seed(seed=999):
    """
    Sets global seeds for testing in numpy, random, and numbaized numpy.
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    return


@fixture(scope="session", name="june_world_path")
def get_june_world_path():
    return Path(__file__).parent / "data/june_world.h5"


@fixture(scope="session", name="sampler")
def make_sampler():
    return TransmissionSampler.from_file()


@fixture(name="agent_data")
def make_agent_data(sampler):
    n_agents = 100
    data = HeteroData()
    data["agent"].id = torch.arange(0, n_agents)
    data["agent"].age = torch.randint(0, 100, (n_agents,))
    data["agent"].sex = torch.randint(0, 2, (n_agents,))
    inf_params = {}
    inf_params_values = sampler(n_agents)
    inf_params["max_infectiousness"] = inf_params_values[0]
    inf_params["shape"] = inf_params_values[1]
    inf_params["rate"] = inf_params_values[2]
    inf_params["shift"] = inf_params_values[3]
    data["agent"].infection_parameters = inf_params
    data["agent"].transmission = torch.zeros(n_agents)
    data["agent"].susceptibility = torch.ones(n_agents)
    data["agent"].is_infected = torch.zeros(n_agents)
    data["agent"].infection_time = torch.zeros(n_agents)
    data["agent"].infected_probs = torch.zeros(n_agents)
    symptoms = {}
    symptoms["current_stage"] = torch.ones(n_agents, dtype=torch.long)
    symptoms["next_stage"] = torch.ones(n_agents, dtype=torch.long)
    symptoms["time_to_next_stage"] = torch.zeros(n_agents)
    data["agent"].symptoms = symptoms
    return data


@fixture(name="data")
def make_data(agent_data):
    data = agent_data
    data["school"].id = torch.arange(0, 4)
    data["school"].people = 25 * torch.ones(4)
    data["company"].id = torch.arange(0, 4)
    data["company"].people = 25 * torch.ones(4)
    # data["leisure"].id = torch.arange(0, 4)
    # data["leisure"].people = 25 * torch.ones(4)
    data["household"].id = torch.arange(0, 25)
    data["household"].people = 4 * torch.ones(25)
    data["agent", "attends_school", "school"].edge_index = torch.vstack(
        (data["agent"].id, torch.tensor(np.repeat(np.arange(0, 4), 25)))
    )
    data["agent", "attends_company", "company"].edge_index = torch.vstack(
        (data["agent"].id, torch.tensor(np.repeat(np.arange(0, 4), 25)))
    )
    data["agent", "attends_household", "household"].edge_index = torch.vstack(
        (data["agent"].id, torch.tensor(np.repeat(np.arange(0, 25), 4)))
    )
    data = T.ToUndirected()(data)
    return data


@fixture(name="inf_data")
def make_inf_data(data):
    return infect_people_at_indices(data, list(range(0, 100, 10)))


@fixture(name="timer")
def make_timer():
    timer = Timer(
        initial_day="2022-02-01",
        total_days=10,
        weekday_step_duration=(8, 8, 8),
        weekend_step_duration=(
            12,
            12,
        ),
        weekday_activities=(
            ("company", "school", "household"),
            ("leisure", "household"),
            ("household",),
        ),
        weekend_activities=(
            ("leisure",),
            ("household",),
        ),
    )
    return timer


@fixture(name="school_timer")
def make_school_timer():
    timer = Timer(
        initial_day="2022-02-01",
        total_days=10,
        weekday_step_duration=(24,),
        weekend_step_duration=(24,),
        weekday_activities=(("school",),),
        weekend_activities=(("school",),),
    )
    return timer



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages, Extension
from setuptools.command.install import install
import subprocess
import os
from os.path import abspath, dirname, join

this_dir = abspath(dirname(__file__))
with open(join(this_dir, "LICENSE")) as f:
    license = f.read()

with open(join(this_dir, "README.md"), encoding="utf-8") as file:
    long_description = file.read()

with open(join(this_dir, "requirements.txt")) as f:
    requirements = f.read().split("\n")


setup(
    name="grad_june",
    version="0.1.0",
    description="Implementation of JUNE using Graph Neural Networks in PyTorch.",
    url="https://github.com/arnauqb/GradABM-JUNE",
    long_description_content_type="text/markdown",
    long_description=long_description,
    author="Arnau Quera-Bofarull",
    author_email="arnauq@protonmail.com",
    license="MIT License",
    install_requires=requirements,
    packages=find_packages(exclude=["docs"]),
    include_package_data=True,
)



import torch
import yaml
from torch.utils.checkpoint import checkpoint

from grad_june import (
    TransmissionUpdater,
    IsInfectedSampler,
    SymptomsUpdater,
    InfectionNetworks,
)
from grad_june.policies import Policies
from grad_june.cuda_utils import get_fraction_gpu_used
from grad_june.paths import default_config_path


class GradJune(torch.nn.Module):
    """
    This class represents an epidemiological simulation model.

    Attributes:
        symptoms_updater: An object that updates agents' symptoms based on their infection status.
        policies: An object that contains the current policies used for the simulation.
        infection_networks: An object that calculates the probability of not being infected for each agent based on current policies.
        transmission_updater: An object that updates agent transmission based on current transmission updater values.
        is_infected_sampler: An object that samples which agents will be infected based on their not_infected probabilities.
        device: A string representing the device being used for the simulation.
    """

    def __init__(
        self,
        symptoms_updater=None,
        policies=None,
        infection_networks=None,
        device="cpu",
    ):
        super().__init__()

        # Initializes symptoms updater, policies, and infection networks.
        if symptoms_updater is None:
            symptoms_updater = SymptomsUpdater.from_file()
        self.symptoms_updater = symptoms_updater
        if policies is None:
            policies = Policies.from_file()
        self.policies = policies
        if infection_networks is None:
            infection_networks = InfectionNetworks.from_file()
        self.infection_networks = infection_networks

        # Initializes transmission updater, is_infected_sampler, and device.
        self.transmission_updater = TransmissionUpdater()
        self.is_infected_sampler = IsInfectedSampler()
        self.device = device

    @classmethod
    def from_file(cls, fpath=default_config_path):
        """
        This method creates a new instance of the GradJune class from a configuration file.

        Args:
            fpath: A string representing the path to the configuration file.

        Returns:
            A new instance of the GradJune class.
        """
        with open(fpath, "r") as f:
            params = yaml.safe_load(f)
        return cls.from_parameters(params)

    @classmethod
    def from_parameters(cls, params):
        """
        This method creates a new instance of the GradJune class from a dictionary of parameters.

        Args:
            params: A dictionary containing the necessary parameters.

        Returns:
            A new instance of the GradJune class.
        """
        symptoms_updater = SymptomsUpdater.from_parameters(params)
        policies = Policies.from_parameters(params)
        infection_networks = InfectionNetworks.from_parameters(params)
        return cls(
            symptoms_updater=symptoms_updater,
            policies=policies,
            infection_networks=infection_networks,
            device=params["system"]["device"],
        )

    def infect_people(self, data, timer, new_infected):
        """
        This method infects people based on the given parameters.

        Args:
            data: A dictionary containing simulation data.
            timer: An integer representing the current simulation time.
            new_infected: A tensor representing which agents will be infected.

        Returns:
            None.
        """
        # Updates agent susceptibility, infection status, and infection time based on new_infected tensor.
        data["agent"].susceptibility = torch.maximum(
            torch.tensor(0.0, device=self.device),
            data["agent"].susceptibility - new_infected,
        )
        data["agent"].is_infected = data["agent"].is_infected + new_infected
        data["agent"].infection_time = data["agent"].infection_time + new_infected * (
            timer.now - data["agent"].infection_time
        )

    def forward(self, data, timer):
        """
        This function represents a forward pass through the epidemiological simulation model.

        Args:
            data: A PyTorch geometric data object containing simulation data.
            timer: An integer representing the current simulation time.

        Returns:
            A A PyTorch geometric data object containing updated simulation data.
        """

        # Updates agent transmission based on current transmission updater values.
        data["agent"].transmission = self.transmission_updater(data=data, timer=timer)

        # Calculates probability of not being infected for each agent based on current policies.
        not_infected_probs = self.infection_networks(
            data=data,
            timer=timer,
            policies=self.policies,
        )

        # Samples which agents will be infected based on their not_infected probabilities.
        new_infected = self.is_infected_sampler(not_infected_probs)

        # Infects agents who were sampled as new_infected.
        self.infect_people(data, timer, new_infected)

        # Updates agents' symptoms based on their infection status.
        self.symptoms_updater(data=data, timer=timer, new_infected=new_infected)

        # Returns updated simulation data.
        return data



import numpy as np


def convert_lognormal_parameters(mean, std):
    """
    Converts mean and std to loc and scale parmaeters for the LogNormal.
    """
    loc = np.log(mean**2 / np.sqrt(mean**2 + std**2))
    scale = np.sqrt(np.log(1 + std**2 / mean**2))
    return loc, scale


def make_parameters():
    ret = {}

    """
    Symptoms Parameters
    Taken from the Covasim paper
    """
    symptoms = {}
    stages = [
        "recovered",
        "susceptible",
        "exposed",
        "infectious",
        "symptomatic",
        "severe",
        "critical",
        "dead",
    ]
    symptoms["stages"] = stages

    # Symptom transition probabilities
    tprobs = {}
    tprobs["recovered"] = {"0-100": 0.0}
    tprobs["susceptible"] = {"0-100": 0.0}
    tprobs["exposed"] = {"0-100": 1.0}
    tprobs["infectious"] = {
        "0-10": 0.5,
        "10-20": 0.55,
        "20-30": 0.6,
        "30-40": 0.65,
        "40-50": 0.7,
        "50-60": 0.75,
        "60-70": 0.8,
        "70-80": 0.85,
        "80-90": 0.9,
        "90-100": 0.9,
    }
    tprobs["symptomatic"] = {
        "0-10": 0.0005,
        "10-20": 0.00165,
        "20-30": 0.00720,
        "30-40": 0.02080,
        "40-50": 0.03430,
        "50-60": 0.07650,
        "60-70": 0.13280,
        "70-80": 0.20655,
        "80-90": 0.24570,
        "90-100": 0.24570,
    }
    tprobs["severe"] = {
        "0-10": 0.00003,
        "10-20": 0.00008,
        "20-30": 0.00036,
        "30-40": 0.00104,
        "40-50": 0.00216,
        "50-60": 0.00933,
        "60-70": 0.03639,
        "70-80": 0.08923,
        "80-90": 0.17420,
        "90-100": 0.17420,
    }
    tprobs["critical"] = {
        "0-10": 0.00002,
        "10-20": 0.00002,
        "20-30": 0.0001,
        "30-40": 0.00032,
        "40-50": 0.00098,
        "50-60": 0.00265,
        "60-70": 0.00766,
        "70-80": 0.02439,
        "80-90": 0.08292,
        "90-100": 0.16190,
    }
    # Convert to relative probs to the previous stage.
    tprobs["critical"] = {
        key: tprobs["critical"][key] / tprobs["severe"][key]
        for key in tprobs["critical"]
    }
    tprobs["severe"] = {
        key: tprobs["severe"][key] / tprobs["symptomatic"][key]
        for key in tprobs["severe"]
    }
    tprobs["symptomatic"] = {
        key: tprobs["symptomatic"][key] / tprobs["infectious"][key]
        for key in tprobs["symptomatic"]
    }
    symptoms["stage_transition_probabilities"] = tprobs

    # Symptom transition times
    ttimes = {}
    loc, scale = convert_lognormal_parameters(4.5, 1.5)
    ttimes["exposed"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(1.1, 0.9)
    ttimes["infectious"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(6.6, 4.9)
    ttimes["symptomatic"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(1.5, 2.0)
    ttimes["severe"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(10.7, 4.8)
    ttimes["critical"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    symptoms["stage_transition_times"] = ttimes

    # Recovery times
    rtimes = {}
    # loc, scale = convert_lognormal_parameters(4.5, 1.5)
    # rtimes["recovered"] = {"dist": "LogNormal", "loc": loc, "scale" : scale}
    # loc, scale = convert_lognormal_parameters(4.5, 1.5)
    # rtimes["susceptible"] = {"dist": "LogNormal", "loc": loc, "scale" : scale}
    loc, scale = convert_lognormal_parameters(4.5, 1.5)
    rtimes["exposed"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(8.0, 2.0)
    rtimes["infectious"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(8.0, 2.0)
    rtimes["symptomatic"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(18.1, 6.3)
    rtimes["severe"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    loc, scale = convert_lognormal_parameters(18.1, 6.3)
    rtimes["critical"] = {"dist": "LogNormal", "loc": loc, "scale": scale}
    symptoms["recovery_times"] = rtimes
    ret["symptoms"] = symptoms

    # policy parameters
    policies = {}
    policies["interaction"] = {
        "social_distancing": {
            1: {
                "start_date": "2020-03-16",
                "end_date": "2020-03-24",
                "beta_factors": {
                    "leisure": 0.65,
                    "care_home": 0.65,
                    "school": 0.65,
                    "university": 0.65,
                    "company": 0.65,
                },
            },
            2: {
                "start_date": "2020-03-24",
                "end_date": "2020-05-11",
                "beta_factors": {
                    "leisure": 0.45,
                    "care_home": 0.45,
                    "school": 0.45,
                    "university": 0.45,
                    "company": 0.45,
                },
            },
            3: {
                "start_date": "2020-05-11",
                "end_date": "2020-07-04",
                "beta_factors": {
                    "leisure": 0.50,
                    "care_home": 0.50,
                    "school": 0.50,
                    "university": 0.50,
                    "company": 0.50,
                },
            },
        }
    }
    policies["close_venue"] = {
        "close_venue": {
            1: {
                "start_date": "2020-03-21",
                "end_date": "2020-07-04",
                "names": ["leisure", "school"],
            },
        }
    }
    policies["quarantine"] = {
        "quarantine": {
            "start_date": "2020-03-16",
            "end_date": "9999-03-24",
            "stage_threshold": 4,
        },
    }
    ret["policies"] = policies

    return ret
